// draw_fig.c

#include <basic.h>
#include <btron/btron.h>
#include <btron/hmi.h>
#include <bstring.h>
#include <btron/libapp.h>

#include "segment.h"
#include "grapherr.h"
#include "debug.h"
#include "graph.h"

#define FPAT(i) ((PAT*)&pattern[(i)])

// UWORD -> UH
typedef struct {
  B type;
  B subid;
  UH id;
  UH hsize;
  UH vsize;
  B mask[32];
} MASKDATA;

// UWORD -> UH
typedef struct {
  B type;
  B subid;
  UH id;
  UH hsize;
  UH vsize;
  UH ncol;
  UH fgcol_lower;
  UH fgcol_upper;
  UH bgcol_lower;
  UH bgcol_upper;
  UH mask;
} PATSEG;

static MASKDATA maskbuff[48];
static PAT pattern[48];

// ---------------------------------------
// Ķν
// ---------------------------------------
void initDrawEnv()
{
  W i;
  W id;

  for (i = 0; i < 32; ++i) {
    if (i >= 1 && i <= 13) {
      id = (i < 7) ? i : 7;
      pattern[i].spat.kind = 0;
      pattern[i].spat.hsize = 16;
      pattern[i].spat.vsize = 16;
      pattern[i].spat.fgcol = 0x10000000;
      pattern[i].spat.bgcol = 0x10ffffff;
      pattern[i].spat.mask = (B*)id;
    }
    else {
      pattern[i].spat.kind = -1;
      pattern[i].spat.mask = (B*)&maskbuff[i];
    }
    maskbuff[i].type = -1;
  }
}

void freeDrawEnv()
{
}

// ---------------------------------------
// ???
// ---------------------------------------
W GetFigRect(FSEGMENT* seg, RECT* rect)
{
  RECT fig;
  W sts = 0;

  // ENTRY("GetFigRect()");

  switch(seg->fig.rect.subid) {
  case 0: // Ĺ
    fig = ((TF_RECT *)&seg->fig)->frame;
    break;
    
  case 1: // ѴĹ
    fig = ((TF_RRECT *)&seg->fig)->frame;
    break;

  case 2: // ʱߡʱߡ
    fig = ((TF_OVAL *)&seg->fig)->frame;
    break;

  case 3: // 
    fig = ((TF_SEC *)&seg->fig)->frame;
    break;

  case 4: // ݷ
    fig = ((TF_CHORD *)&seg->fig)->frame;
    break;

  case 6: // ľ
    fig.c.left = ((TF_LINE *)&seg->fig)->start.x;
    fig.c.top = ((TF_LINE *)&seg->fig)->start.y;
    fig.c.right = ((TF_LINE *)&seg->fig)->end.x;
    fig.c.bottom = ((TF_LINE *)&seg->fig)->end.y;
    if (fig.c.top == fig.c.bottom)
      fig.c.bottom++;
    if (fig.c.left == fig.c.right)
      fig.c.right++;
    AdjustRect(&fig);
    break;
    
  case 7: // ʱ߸̡ʱ߸̡
    *rect = ((TF_ARC *)&seg->fig)->frame;
    break;

  case 5: /* ¿ѷ */
  case 8: /* ޤ */
  case 9: /*  */
  case 10: /* ޡ */
  case 11: /* Ǥտ޷ */
  default:
    nullrect(fig);
    sts = EA_NOSPT;
    break;
  }

 exit:
  *rect = fig;
  if (sts < 0) {
    ERROR(sts, "GetFigRect[1]");
  }
  // EXIT("GetFigRect()");

  return sts;
}

// ===============================================
//  ޷ǥ
// ===============================================

// ---------------------------------------
// Ĺ
// ---------------------------------------
W DrawRect(W gid, TF_RECT* rec)
{
  W sts;

  // ENTRY("DrawRect()");

  if (rec->f_pat != 0) {
    sts = gfil_rec(gid, rec->frame, FPAT(rec->f_pat), rec->angle, rec->mode);
    if (sts < 0) {
      ERROR(sts, "DrawRect[1]");
      return sts;
    }
  }
  if (rec->l_pat != 0) {
    sts = gfra_rec(gid, rec->frame, rec->l_atr, FPAT(rec->l_pat), rec->angle, rec->mode);
    if (sts < 0) {
      ERROR(sts, "DrawRect[2]");
      return sts;
    }
  }
  // EXIT("DrawRect()");

  return 0;
}

// ---------------------------------------
// ѴĹ
// ---------------------------------------
W DrawRrect(W gid, TF_RRECT* rrc)
{
  W sts;

  if (rrc->f_pat != 0) {
    sts = gfil_rrc(gid, rrc->frame, rrc->rh, rrc->rv, FPAT(rrc->f_pat),
                   rrc->angle, rrc->mode);
    if (sts < 0) {
      ERROR(sts, "DrawRrect[1]");
      return sts;
    }
  }
  if (rrc->l_pat != 0) {
    sts = gfra_rrc(gid, rrc->frame, rrc->rh, rrc->rv, rrc->l_atr, 
                   FPAT(rrc->l_pat), rrc->angle, rrc->mode);
    if (sts < 0) {
      ERROR(sts, "DrawRrect[2]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ʱߡʱߡ
// ---------------------------------------
W DrawOval(W gid, TF_OVAL* ovl)
{
  W sts;

  if (ovl->f_pat != 0) {
    sts = gfil_ovl(gid, ovl->frame, FPAT(ovl->f_pat),
                   ovl->angle, ovl->mode);
    if (sts < 0) {
      ERROR(sts, "DrawOval[1]");
      return sts;
    }
  }
  if (ovl->l_pat != 0) {
    sts = gfra_ovl(gid, ovl->frame, ovl->l_atr, FPAT(ovl->l_pat),
                   ovl->angle, ovl->mode);
    if (sts < 0) {
      ERROR(sts, "DrawOval[2]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// 
// ---------------------------------------
W DrawSec(W gid, TF_SEC* sec)
{
  W sts;

  if (sec->f_pat != 0) {
    sts = gfil_sec(gid, sec->frame, sec->start, sec->end, FPAT(sec->f_pat),
                   sec->angle, sec->mode);
    if (sts < 0) {
      ERROR(sts, "DrawSec[1]");
      return sts;
    }
  }
  if (sec->l_pat != 0) {
    sts = gfra_sec(gid, sec->frame, sec->start, sec->end, sec->l_atr,
                   FPAT(sec->l_pat), sec->angle, sec->mode);
    if (sts < 0) {
      ERROR(sts, "DrawSec[2]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ݷ
// ---------------------------------------
W DrawChord(W gid, TF_CHORD* cho)
{
  W sts;

  if (cho->f_pat != 0) {
    sts = gfil_cho(gid, cho->frame, cho->start, cho->end, FPAT(cho->f_pat),
                   cho->angle, cho->mode);
    if (sts < 0) {
      ERROR(sts, "DrawChord[1]");
      return sts;
    }
  }
  if (cho->l_pat != 0) {
    sts = gfra_cho(gid, cho->frame, cho->start, cho->end, cho->l_atr,
                   FPAT(cho->l_pat), cho->angle, cho->mode);
    if (sts < 0) {
      ERROR(sts, "DrawChord[2]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ¿ѷ
// ---------------------------------------
W DrawPolygon(W gid, TF_POLYGON* pol)
{
  W sts;

  // 2nd arg -> POLY
  if (pol->f_pat != 0) {
    sts = gfil_pol(gid, (POLY*)&pol->round, FPAT(pol->f_pat), pol->mode);
    if (sts < 0) {
      ERROR(sts, "DrawPolygon[1]");
      return sts;
    }
  }
  if (pol->l_pat != 0) {
    sts = gfra_pol(gid, (POLY*)&pol->round, pol->l_atr, FPAT(pol->l_pat), pol->mode);
    if (sts < 0) {
      ERROR(sts, "DrawPolygon[2]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ľ
// ---------------------------------------
W DrawLine(W gid, TF_LINE* lin)
{
  W sts;

  // ENTRY("DrawLine()");

  if (lin->l_pat != 0) {
    sts = gdra_lin(gid, lin->start, lin->end, lin->l_atr, FPAT(lin->l_pat),
                   lin->mode);
    if (sts < 0) {
      ERROR(sts, "DrawLine[1]");
      return sts;
    }
  }
  // EXIT("DrawLine()");

  return 0;
}

// ---------------------------------------
// ʱ߸̡ʱ߸̡
// ---------------------------------------
W DrawArc(W gid, TF_ARC* arc)
{
  W sts;

  if (arc->l_pat != 0) {
    sts = gdra_arc(gid, arc->frame, arc->start, arc->end, arc->l_atr,
                   FPAT(arc->l_pat), arc->angle, arc->mode);
    if (sts < 0) {
      ERROR(sts, "DrawArc[1]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ޤ
// ---------------------------------------
W DrawPolyline(W gid, TF_POLYLINE* pln)
{
  W sts;

  // ENTRY("DrawPolyline()");

  if (pln->l_pat != 0) {
    sts = gdra_pln(gid, (POLY*)&pln->round, pln->l_atr, FPAT(pln->l_pat),
                   pln->mode);
    if (sts < 0) {
      ERROR(sts, "DrawPolyline[1]");
      return sts;
    }
  }
  // EXIT("DrawPolyline()");

  return 0;
}

// ---------------------------------------
// 
// ---------------------------------------
W DrawSpline(W gid, TF_SPLINE* spl)
{
  W sts;

  if (spl->l_pat != 0) {
    sts = gdra_spl(gid, spl->np, spl->pt, spl->l_atr, FPAT(spl->l_pat),
                   spl->mode);
    if (sts < 0) {
      ERROR(sts, "DrawSpline[1]");
      return sts;
    }
  }
  return 0;
}

// ---------------------------------------
// ޡ
// ---------------------------------------
W DrawMarker(W gid, TF_MARKER* mrk)
{
  W sts;

  return EA_NOSPT;
}

// ---------------------------------------
// Ǥտ޷
// ---------------------------------------
W DrawScan(W gid, TF_SCAN* scn)
{
  W sts, nh, i, j, tmpx;
  W akey = -1;
  W size;
  GRGN* gscan = NULL;

  if (scn->f_pat != 0) {
    size = sizeof(GRGN) + (sizeof(W) * 2 * (scn->nr - 1));
    
    // ͭΰƤꤹ
    gscan = malloc(size);
    if (gscan == 0) {
        ERROR(sts, "DrawScan[1]");
        return 0;
    }

    // --- CAUTION !!! 
    // --- below lines are not sure (yashiro)
    
    gscan->r.c.top = scn->sy;
    gscan->r.c.bottom = scn->sy + scn->nr - 1;
    gscan->ns = scn->nr;
    nh = 0;
    for (i = 0; i < scn->nr; ++i){
      gscan->s[i].y = scn->sy + i;
      gscan->s[i].hp = (HRGN*)((B*)&scn[1] + nh);
      for (j = 0; j < gscan->s[i].hp->nx; ++j){
        tmpx = gscan->s[i].hp->x[j];
        if (gscan->r.c.left < tmpx || i == 0){
          gscan->r.c.left = tmpx;
        }
        if (gscan->r.c.right > tmpx || i == 0){
          gscan->r.c.right = tmpx;
        }
      }
      nh += sizeof(UH) * (1 + gscan->s[i].hp->nx);
    }
    
    // --- CAUTION !!! 
    // --- above lines are not sure (yashiro)
    
    if ((sts = gfil_gen(gid, gscan, FPAT(scn->f_pat), scn->mode)) < 0) {
      ERROR(sts, "DrawScan[3]");
    }
  }
 exit:
  if (gscan != NULL) {
      free(gscan);
  }
  return sts;
}

W DrawFig(W gid, FSEGMENT* seg)
{
  // ENTRY("DrawFig()");

  if (SegmentType(seg->tad.id) != SEG_ESC) {
    ERROR(EA_PAR, "DrawFig[1]");
    return EA_PAR;
  }
  if ((seg->tad.id & 0x00ff) != TS_FPRIM) {
    ERROR(EA_PAR, "DrawFig[2]");
    return EA_PAR;
  }
  switch(seg->fig.rect.subid) {
  case 0: // Ĺ
    return DrawRect(gid, &seg->fig.rect);
    
  case 1: // ѴĹ
    return DrawRrect(gid, &seg->fig.rrect);
    
  case 2: // ʱߡʱߡ
    return DrawOval(gid, &seg->fig.oval);
    
  case 3: // 
    return DrawSec(gid, &seg->fig.sec);
    
  case 4: // ݷ
    return DrawChord(gid, &seg->fig.chord);
    
  case 5: // ¿ѷ
    return DrawPolygon(gid, &seg->fig.polygon);
    
  case 6: // ľ
    return DrawLine(gid, &seg->fig.line);
    
  case 7: // ʱ߸̡ʱ߸̡
    return DrawArc(gid, &seg->fig.arc);
    
  case 8: // ޤ
    return DrawPolyline(gid, &seg->fig.polyline);
    
  case 9: // 
    return DrawSpline(gid, &seg->fig.spline);
    
  case 10: // ޡ
    return DrawMarker(gid, &seg->fig.marker);
    
  case 11: // Ǥտ޷
    return DrawScan(gid, &seg->fig.scan);
  }
  ERROR(EA_NOSPT, "DrawFig[3]");
  return EA_NOSPT;
}
  
// ===============================================
//  ǡ
// ===============================================

// ---------------------------------------
// 顼ޥå
// ---------------------------------------
W DrawDefColmap(W gid, TF_COLMAP* cmap)
{
  W sts;
  if ((sts = gset_col(gid, 0, (COLOR*)cmap->col, cmap->nent)) < 0) {
    ERROR(sts, "DrawDefColmap[1]");
    return sts;
  }
  return 0;
}

// ---------------------------------------
// ޥǡ
// ---------------------------------------
W DrawDefMask(W gid, TF_MASK* mask)
{
  W id;

  // ENTRY("DrawDefMask()");

  if (mask->type != 0) {
    ERROR(EA_NOSPT, "DrawDefMask[1]");
    return EA_NOSPT;
  }
  if ((id = mask->id) >= 32) {
    ERROR(EA_NOSPT, "DrawDefMask[2]");
    return EA_NOSPT;
  }
  if (mask->hsize != 16 || mask->vsize != 16) {
    ERROR(EA_NOSPT, "DrawDefMask[3]");
    return EA_NOSPT;
  }
  
  memcpy(&maskbuff[id], mask, sizeof(MASKDATA));

  // EXIT("DrawDefMask()");

  return 0;
}

// ---------------------------------------
// ѥ
// ---------------------------------------
W DrawDefPattern(W gid, PATSEG* pat)
{
  W pid, mid;
  B* mask;

  // ENTRY("DrawDefPattern()");

  if (pat->type != 0) {
    ERROR(EA_NOSPT, "DrawDefPattern[1]");
    return EA_NOSPT;
  }
  if ((pid = pat->id) >= 32) {
    ERROR(EA_NOSPT, "DrawDefPattern[2]");
    return EA_NOSPT;
  }
  if (pat->hsize != 16 || pat->vsize != 16) {
    ERROR(EA_NOSPT, "DrawDefPattern[3]");
    return EA_NOSPT;
  }
  if (pat->ncol != 1) {
    ERROR(EA_NOSPT, "DrawDefPattern[4]");
    return EA_NOSPT;
  }
  mid = pat->mask;
  if (maskbuff[mid].type == -1) {
    ERROR(EA_MASK, "DrawDefPattern[4]");
    return EA_MASK;
  }
  pattern[pid].spat.kind = 0;
  pattern[pid].spat.hsize = pat->hsize;
  pattern[pid].spat.vsize = pat->vsize;
  pattern[pid].spat.fgcol =
    (pat->fgcol_upper << 16) | (pat->fgcol_lower & 0xffff);
  pattern[pid].spat.bgcol =
    (pat->bgcol_upper << 16) | (pat->bgcol_lower & 0xffff);
  if (mid < 14) {
    pattern[pid].spat.mask = (B*)mid;
  }
  else {
//    pattern[pid].spat.mask = maskbuff[18].mask;
    pattern[pid].spat.mask = maskbuff[mid].mask;
  }
  // EXIT("DrawDefPattern()");

  return 0;
}

// ---------------------------------------
// 
// ---------------------------------------
W DrawDefLkind(W gid, TF_LKIND* lkind)
{
  return EA_NOSPT;
}

// ---------------------------------------
// ޡ
// ---------------------------------------
W DrawDefMark(W gid, TF_MARKPAT* mark)
{
  return EA_NOSPT;
}

// ---------------------------------------
// ???
// ---------------------------------------
W DrawDefine(W gid, DSEGMENT* seg)
{
  // ENTRY("DrawDefine()");

  if (SegmentType(seg->tad.id) != SEG_ESC) {
    ERROR(EA_PAR, "DrawDefine[1]");
    return EA_PAR;
  }
  if ((seg->tad.id & 0x00ff) != TS_FDEF) {
    ERROR(EA_PAR, "DrawDefine[2]");
    return EA_PAR;
  }
  switch(seg->def.cmap.subid) {
  case 0: // 顼ޥå
    return DrawDefColmap(gid, &seg->def.cmap);
    
  case 1: // ޥǡ
    return DrawDefMask(gid, &seg->def.mask);
    
  case 2: // ѥ
    return DrawDefPattern(gid, (PATSEG*)&seg->def.pat);
    
  case 3: // 
    return DrawDefLkind(gid, &seg->def.lkind);
    
  case 4: // ޡ
    return DrawDefMark(gid, &seg->def.mark);
  }
  ERROR(EA_NOSPT, "DrawDefine[3]");
  return EA_NOSPT;
}
  
// ===============================================
//  롼
// ===============================================
W DrawGroup(W gid, GSEGMENT* seg)
{
  if (SegmentType(seg->tad.id) != SEG_ESC) {
    ERROR(EA_PAR, "DrawGroup[1]");
    return EA_PAR;
  }
  if ((seg->tad.id & 0x00ff) != TS_FGRP) {
    ERROR(EA_PAR, "DrawGroup[2]");
    return EA_PAR;
  }
  switch(seg->grp.start.subid) {
  case 0: // 롼׳
    return 0; // ̵뤹
    
  case 1: // 롼׽λ
    return 0; // ̵뤹
  }
  ERROR(EA_NOSPT, "DrawGroup[2]");
  return EA_NOSPT;
}
  
// ===============================================
//  ???
// ===============================================
W Draw(W gid, RECT clip, TADSEG* seg)
{
  RECT fig;
  W sts = EA_NOSPT;

  // ENTRY("Draw()");

  if (SegmentType(seg->id) != SEG_ESC) {
    sts = EA_PAR;
    goto exit;
  }
  switch (seg->id & 0x00ff) {
  case TS_FPRIM: // ޷ǥ
    GetFigRect((FSEGMENT*)seg, &fig);
    if (sectrect(clip, fig)) {
      sts = DrawFig(gid, (FSEGMENT*)seg);
    }
    else {
      sts = 0;
    }
    break;

  case TS_FDEF:  // ǡ
    sts = DrawDefine(gid, (DSEGMENT*)seg);
    break;

  case TS_FGRP:  // 롼
    sts = DrawGroup(gid, (GSEGMENT*)seg);
    break;

  case TS_FMAC:  // ޥȥ
  case TS_FATTR: // ޷
    // case TS_FPAGE: // ޷ڡդ
  case TS_FMEMO: // ޷
  case TS_FAPPL: // ޷ץꥱ
    sts = EA_NOSPT;
    ERROR(sts, "Draw[1]");
  }

 exit:
  // EXIT("Draw()");

  return sts;
}

